#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include <memory.h>
#include <tchar.h>

#include <Python.h>

static PyObject *traceback_module;
static PyObject *string_module;

static PyObject *winmain_func;

static PyObject *import_module(char *name)
{
	PyObject *str, *module;
	char msg[128];

	str = PyString_FromString(name);
	if (str == NULL)
		return NULL;
	module = PyImport_Import(str);
	Py_DECREF(str);
	if (module == NULL) {
		sprintf(msg, "Could not import %s module", name);
		MessageBox(NULL, msg, "Python Error", MB_ICONEXCLAMATION | MB_OK | MB_TASKMODAL);
		return NULL;
	}

	return module;
}

static void python_error()
{
	PyObject *exc = NULL, *value = NULL, *trace = NULL,
		*func = NULL,
		*lines = NULL,
		*str = NULL,
		*obj = NULL;

	// exc, value, trace = sys.exc_info()
	PyErr_Fetch(&exc, &value, &trace);

	// import traceback
	if (traceback_module == NULL)
		traceback_module = import_module("traceback");
	if (traceback_module == NULL)
		goto error;

	// lines = traceback.format_exception(exc, value, trace, 20)
	obj = PyModule_GetDict(traceback_module);
	func = PyDict_GetItemString(obj, "format_exception");
	Py_DECREF(obj); obj = NULL;
	if (func == NULL)
		goto error;
	lines = PyObject_CallFunction(func, "OOOi", exc, value, trace, 20);
	if (lines == NULL)
		goto error;
	Py_DECREF(func); func = NULL;

	// import string
	if (string_module == NULL)
		string_module = import_module("string");
	if (string_module == NULL)
		goto error;

	// str = string.join(lines)
	obj = PyModule_GetDict(string_module);
	func = PyDict_GetItemString(obj, "join");
	Py_DECREF(obj); obj = NULL;
	if (func == NULL)
		goto error;
	str = PyObject_CallFunction(func, "O", lines);
	if (str == NULL)
		goto error;
	Py_DECREF(func); func = NULL;

	MessageBox(NULL, PyString_AsString(str), "Python Error", MB_ICONEXCLAMATION | MB_OK | MB_TASKMODAL);

error:
	Py_XDECREF(exc); Py_XDECREF(value); Py_XDECREF(trace);
	Py_XDECREF(func);
	Py_XDECREF(lines);
	Py_XDECREF(str);
	Py_XDECREF(obj);

	PyErr_Clear();
}

static void python_init_module(char *name)
{
	PyObject *obj = NULL, *mod = NULL, *dict = NULL;

	Py_Initialize();

	obj = PyString_FromString(name);
	if (obj != NULL) {
		mod = PyImport_Import(obj);
		if (PyErr_Occurred())
			mod = NULL;
	}
	if (mod != NULL)
		dict = PyModule_GetDict(mod);
	if (dict != NULL)
		winmain_func = PyDict_GetItemString(dict, "winmain");

	Py_XDECREF(obj);
	Py_XDECREF(mod);
	Py_XDECREF(dict);
}

int APIENTRY WinMain(HINSTANCE inst, HINSTANCE prev_inst, LPSTR cmd_line, int cmd_show)
{
	PyObject *obj;
	int ret;

	python_init_module(cmd_line);
	if (winmain_func == NULL) {
		python_error();
		return 0;
	}

	ret = 0;
	obj = PyObject_CallFunction(winmain_func, "iisi", inst, prev_inst, cmd_line, cmd_show);
	if (obj != NULL)
		ret = PyInt_AsLong(obj);
	Py_XDECREF(obj);
	if (PyErr_Occurred())
		python_error();

	return ret;
}
